@@ -13,16 +13,30 @@ and documentation.
 
 project home: http://code.google.com/p/wyckedsceptre/
 Last Changed Date: $LastChangedDate: 2010-10-26 22:37:34 -0500 (Tue, 26 Oct 2$
-Revision         : $Rev: 36 $
+Revision         : $Rev: 50 $
 """
-import random
-import pickle
-from wyckedsceptre_constants import XP_LEVELS
+import wyckedsceptre_constants
+
+
+def assert_positive_integer(val_to_test, val_name='Value'):
+    """
+    Verifies that a value is a positive integer or zero.
+    @param val_to_test Value to test if it is a positive integer or zero
+    @param val_name A string name to describe that variable, to be used
+            in the exception text if raised
+    @except ValueError If val_to_test is < 0 or not an integer
+    """
+    if not isinstance(val_to_test, int) or val_to_test < 0:
+        raise ValueError(val_name + ' must be a positive integer or zero')
 
 
 class Die(object):
     """
     Object for simulating a set of an m-numbered set of n-sided dice
+
+    >>> print wyckedsceptre.Die('1d20')
+    1d20
+
     """
 
     def __init__(self, die_str='1d20'):
@@ -30,12 +44,39 @@ class Die(object):
         Constructor for the dice object
         @param die_str string that takes the form "MdN" where M is the number
         of dice and N is the number of sides each die has
+        @todo add regular expression support die_str
         """
+        if not isinstance(die_str, str):
+            raise TypeError('Die constructor must take a string')
         dstr = die_str.split('d')
         ## the number of dice in a Die object (integer)
         self.num_die = int(dstr[0])
         ## the number of sides on a Die object (integer)
         self.die_sides = int(dstr[1])
+        self.__invariant__()
+
+    def __eq__(self, other):
+        """
+        overloaded equality test for Die object
+        """
+        if isinstance(other, Die):
+            return ((self.num_die == other.num_die) and \
+                    (self.die_sides == other.die_sides))
+        else:
+            return False
+
+    def __invariant__(self):
+        """
+        Tests to insure the Die object's state is valid.
+        Throws a ValueError exception if any problems are detected.
+        @exception ValueError if # of sides or dice is invalid
+        """
+        if not (self.die_sides in wyckedsceptre_constants.VALID_DIE_SIDES):
+            raise ValueError('Number of die sides must be either ' + \
+                    str(wyckedsceptre_constants.VALID_DIE_SIDES))
+        assert_positive_integer(self.num_die, 'Number of Dice')
+        if 0 == self.num_die:
+            raise ValueError('Number of die cannot be zero')
 
     def __str__(self):
         """
@@ -65,6 +106,7 @@ class Die(object):
         Each die is rolled invididually
         @returns the sum of all die rolls as an integer
         """
+        import random
         total = 0
         for _ in range(self.num_die):
             total += random.randint(1, self.die_sides)
@@ -92,6 +134,17 @@ class Attack(object):
         ## a Die object that represents the attack (ex: 1d6)
         self._damage_die = Die(die_roll)
 
+    def __eq__(self, other):
+        """
+        overloaded equality operator for Attack object
+        """
+        if isinstance(other, Attack):
+            return ((self._desc == other._desc)
+                and (self._bonus == other._bonus)
+                and (self._damage_die == other._damage_die))
+        else:
+            return False
+
     def highest(self):
         """
         The highest possible outcome of an attack roll
@@ -121,7 +174,7 @@ def abil_to_mod(ability_score):
     @returns the corresponding modifier for an ability score as an integer
     example: a Strength of 17 has a +3 strength modifier
     """
-    return (ability_score - 10) / 2
+    return (ability_score - 10) // 2
 
 
 class Character(object):
@@ -157,9 +210,9 @@ class Character(object):
         # self.inventory = list(items())
         ## Character's money in Gold Pieces (integer)
         self.money = 100
-        ## Character's Experience xp (integer)
+        ## Character's Experience (integer)
         self.xp = 0
-        ## Bool - is this character an Non-Player Character (NPC)
+        ## Bool - is this character a Non-Player Character (NPC)
         self.is_npc = True
         ## This character's attack (Attack object)
         self.attack = Attack()
@@ -171,22 +224,75 @@ class Character(object):
                             'PARALYSIS-TURNSTONE': 1, 'DRAGONBREATH': 1,
                             'RODS-STAVES-SPELLS': 1}
         ## Character's state (integer between -1 and 1)
-        self.state = 0
+        self.state = 0.0
         self.update_state()
 
+    def __eq__(self, other):
+        """
+        overloaded equality tester for Character object
+        """
+        if isinstance(other, Character):
+            return ((self.name == other.name)
+                # and (self.student_number == other.student_number)
+                and (self.name == other.name)
+                and (self.dnd_class == other.dnd_class)
+                and (self.level == other.level)
+                and (self.hit_points == other.hit_points)
+                and (self.hit_points_max == other.hit_points_max)
+                and (self.armor_class == other.armor_class)
+                and (self.initiative_bonus == other.initiative_bonus)
+                and (self.initiative_total == other.initiative_total)
+                and (self.money == other.money)
+                and (self.xp == other.xp)
+                and (self.is_npc == other.is_npc)
+                and (self.attack == other.attack)
+                and (self.abilities == other.abilities)
+                and (self.saving_throws == other.saving_throws)
+                and (self.state == other.state))
+        else:
+            return False
+
+    def __invariant__(self):
+        """
+        Tests to insure the Character object's state is valid.
+        Throws a ValueError exception if any problems are detected.
+        @exception ValueError a property is not within a valid range
+        @exception TypeError a property is not the right data type
+        """
+        assert_positive_integer(self.xp, 'Character - XP')
+        assert_positive_integer(self.armor_class, 'Character - Armor class')
+        assert_positive_integer(self.money, 'Character - Money')
+        assert_positive_integer(self.initiative_bonus, 'Initiative bonus')
+        assert_positive_integer(self.initiative_total, 'Initiative total')
+        assert_positive_integer(self.hit_points_max, 'Char - max hit points')
+        assert_positive_integer(self.level, 'Character - level')
+        if not isinstance(self.name, str):
+            raise TypeError('Character name must be a string')
+        if not isinstance(self.dnd_class, str):
+            raise TypeError('Character class must be a string')
+        if not isinstance(self.state, float):
+            raise TypeError('Character state must be a float')
+        if self.state > 1.0:
+            raise TypeError('Character state must be less than 1, overhealed')
+
     def __str__(self):
         """
         String conversion routine
         @returns the character's name and a few stats on them
         """
-        res = 'Name: ' + self.name + '\n' + \
-              str(self.abilities) + '\n' + \
-              'hit_points = ' + str(self.hit_points) + '   ' + \
-              'hit_points_max = ' + \
-              str(self.hit_points_max) + '\n' + \
-              'State = ' + str(self.state)
+        res = self.name + ' the level ' + str(self.level) + ' ' + \
+                self.dnd_class + ' HP: ' + str(self.hit_points) + '/' + \
+                str(self.hit_points_max)
         return res
 
+    def get_hp(self):
+        """
+        The current hit points of a Character
+        @returns Current HP of a Character (integer)
+        """
+        self.__invariant__()
+        return self.hit_points
+
     def take_damage(self, amount):
         """
         Reduces the hit_points of a character by "amount".  Amount must be
@@ -196,9 +302,11 @@ class Character(object):
         @exception If amount of damage is less than 1
         @returns the state after damage as a float [-1, 1]
         """
-        assert amount >= 1
+        self.__invariant__()
+        assert_positive_integer(amount, 'Damage')
         self.hit_points -= amount
         self.update_state()
+        self.__invariant__()
         return self.state
 
     def level_at_xp(self):
@@ -206,8 +314,9 @@ class Character(object):
         Computes the level of a character at their current XP level
         @return the level that corresponds to a given XP level
         """
-        ## import math
-        assert self.xp >= 0
+        from wyckedsceptre_constants import XP_LEVELS
+        assert_positive_integer(self.xp, 'Experience')
+        self.__invariant__()
         curr_level = 1
         over = False
         while not over:
@@ -217,22 +326,22 @@ class Character(object):
                 over = True
         return curr_level
 
-    def heal(self, hit_pointsheal):
+    def heal(self, hit_points_to_heal):
         """
         Increases a character's hit points
-        @param hit_pointsheal Number of hit points to restore to a character
-        heal will not heal characters who are completely dead.
-        Will throw an exception if a heal is 0 or less
-        Will only heal a character up to it's maximum hit_points
-        @return the character's new hit_points
-        @exception if hit_pointsheal is 0 or less
+        @param hit_points_to_heal Number of hit points to restore to a
+        Character heal. Will not heal a Character who is completely dead.
+        Will only heal a Character up to it's maximum hit_points
+        @return the Character's new hit_points
+        @exception if hit_points_to_heal is 0 or less
         """
-        assert hit_pointsheal > 0
-        if self.hit_points > -10: # won't heal totally dead people
+        assert_positive_integer(hit_points_to_heal, 'Heal points')
+        self.__invariant__()
+        if self.get_hp() > -10: # won't heal totally dead people
             self.hit_points = min(self.hit_points_max,
-                                    self.hit_points + hit_pointsheal)
+                                    self.get_hp() + hit_points_to_heal)
             self.update_state()
-        return self.hit_points
+        return self.get_hp()
 
     def is_concious(self):
         """
@@ -250,9 +359,11 @@ class Character(object):
         @exception if Character's current hit_points is greater than its max
                     hit_points
         """
-        assert self.hit_points <= self.hit_points_max
-        self.state = self.hit_points / (self.hit_points_max * 1.0)
-        self.level_up()
+        self.__invariant__()
+        # force true division
+        self.state = self.get_hp() / float(self.hit_points_max)
+        self._level_up_()
+        self.__invariant__()
         return self.state
 
     def get_ac(self):
@@ -260,14 +371,16 @@ class Character(object):
         Get the Armor Class of a character
         @return the armor class for a character as an integer
         """
+        assert_positive_integer(self.armor_class, 'Armor Class')
         return self.armor_class
 
-    def level_up(self):
+    def _level_up_(self):
         """
         Sets the character's level according to their XP
         @return true if the level changed
         @return false if no level change occured
         """
+        self.__invariant__()
         new_level = self.level_at_xp()
         if new_level != self.level:
             self.level = new_level
@@ -282,7 +395,10 @@ class Character(object):
         @returns a 1d20 + initiative bonus roll for a character's initiative
         @todo set the total_initiative member
         """
-        return self.initiative_bonus + Die('1d20').roll()
+        self.__invariant__()
+        self.initiative_total = self.initiative_bonus + Die('1d20').roll()
+        self.__invariant__()
+        return self.initiative_total
 
     def get_mod(self, ability):
         """
@@ -433,6 +549,7 @@ def load_file(filename):
     @returns False if the object/file could not be read correctly
     """
     try:
+        import pickle
         file_handler = open(filename, 'r')
     except IOError:
         pass
@@ -456,6 +573,7 @@ def save_file(obj, filename):
     # ex: "bob.char.p" ; "forrest_fight.enc.p"
     # http://wiki.python.org/moin/UsingPickle
     try:
+        import pickle
         res = False
         file_handler = open(filename, 'w')
         pickle.dump(obj, file_handler)
@@ -465,51 +583,3 @@ def save_file(obj, filename):
     finally:
         file_handler.close()
         return res
-
-
-def main():
-    """
-    Test script to simulate a combat between 2 characters
-    """
-#    random.seed(23423)
-
-    mace = Attack('mace', '1d6', 1)
-    print mace.roll_damage()
-
-    bob = roll_new_character('bob', 'magicuser')
-    print str(bob)
-    bob.attack = mace
-
-    for _ in range(10):
-        print bob.make_attack()
-
-    save_file(bob, 'bob.pickle')
-    bob2 = load_file('bob.pickle')
-    print bob2
-
-    # Encounter hit back & forth
-    player = roll_new_character('Player', 'fighter')
-    player.is_npc = False
-    npc = roll_new_character('NPC', 'magicuser')
-    num_rounds = 0
-
-    while player.hit_points > 0 and npc.hit_points > 0:
-        print player.name + ' attacks ' + npc.name
-        [hit_roll, damage] = player.make_attack()
-        if  hit_roll >= npc.get_ac():
-            print player.name + ' attack on ' + npc.name + ' HIT (' + \
-                str(hit_roll) + 'v' + str(npc.get_ac()) + ') & did ' + \
-                str(damage) + ' damage'
-            npc.take_damage(damage)
-        else:
-            print player.name + ' attack on ' + npc.name + ' MISSED (' + \
-                str(hit_roll) + 'v' + str(npc.get_ac()) + ')'
-        print npc
-        num_rounds += 1
-    print 'sample encounter took ' + str(num_rounds) + ' rounds'
-    print 'main() end'
-    return 0
-
-
-if __name__ == '__main__':
-    main()
